/***************************************************************************
                          soundsystem.cpp  -  description
                             -------------------
    begin                : Mon Feb 21 2000
    copyright            : (C) 2000 by Flemming Frandsen
    email                : dion@swamp.dk
 ***************************************************************************/


#include "fluff.h"
#include "soundsystem.h"
#include "sound.h"
#include "sounds.h"
#include "sound_event.h"
#include "map"
#include "loader.h"
#include "game_client.h"
#include "world.h"
#include "obj_base.h"
#include "viewport.h"
#include "SDL_mixer.h"

//-------------------------------------------------------------------
Soundsystem::Soundsystem(Loader *loader_, Game_client *game_client_)
{
	next_sound_handle = 1;
	sound_events_handle = new Sound_events_handle;
	sound_handles_server = new Sound_handles_server;
	loader = loader_;

	int frequency;
	Uint16 format;
	quietmode = Mix_QuerySpec(&frequency,&format,&channels) < 1;
	if (quietmode) {
		logmsg(lt_debug,"Sound system initialized in quiet mode.");
		game_client = NULL;
	} else {
		logmsg(lt_debug,"Sound system initialized, mixing %i channels at %iHz.",channels,frequency);
		game_client = game_client_;
	}
}

//-------------------------------------------------------------------
Soundsystem::~Soundsystem()
{
	if (!quietmode)
		Mix_CloseAudio();

	Sound_events_handle::iterator i = sound_events_handle->begin();
	while (i != sound_events_handle->end())
		delete (i++)->second;
	
	delete sound_events_handle;
	delete sound_handles_server;
}

//-------------------------------------------------------------------
void Soundsystem::update()
{
	Sound_events_handle::iterator i=sound_events_handle->begin();
	struct timeval now;
	gettimeofday(&now,NULL);
	while (i != sound_events_handle->end()) {
		if (!(i->second->flags & SOUND_DONE))
			i->second->check_timer(now);
		if (i->second->flags & SOUND_DONE) {
			Sound_handles_server::iterator k;

			// Can't we just search for a value here?
			for (k = sound_handles_server->begin();
				k != sound_handles_server->end() && k->second != i->first;
				k++);
			if (k != sound_handles_server->end())
				sound_handles_server->erase(k);

			Sound_events_handle::iterator j = i++;
			delete j->second;
			sound_events_handle->erase(j);
		} else {
			if (!quietmode && game_client != NULL)
				i->second->update(game_client);
			i++;
		}
	}
}

//-------------------------------------------------------------------
void Soundsystem::server_think(Serializer_reader *serializer_reader)
{
	Sound_events_handle::iterator i=sound_events_handle->begin();
	while (i != sound_events_handle->end()) {
		if (i->second->flags & SOUND_DIRTY) {
			i->second->flags &= ~SOUND_DIRTY;
			if (i->second->flags & SOUND_DONE) {
				serializer_reader->wint(TT_SOUND_STOP);
				serializer_reader->wint(i->first);
			} else {
				serializer_reader->wint(TT_SOUND_START);
				serializer_reader->wint(i->first);
				i->second->send_event(serializer_reader);
			}
		}
		i++;
	}

	update();
}

//-------------------------------------------------------------------
void Soundsystem::client_think(Serializer_writer &serializer_writer)
{
	while (1) {
		if (serializer_writer.next_int() == TT_SOUND_START) {
			serializer_writer.eat_int(TT_SOUND_START);
			int server_handle = serializer_writer.rint();
			int name_len = serializer_writer.rint()+1;
			char *sample_name = new char[name_len];
			serializer_writer.rwstr(sample_name,name_len);
			float volume = serializer_writer.rfloat();
			int flags = serializer_writer.rint();
			float x_pos = 0;
			float y_pos = 0;
			int objectid = 0;
			if (flags & SOUND_OBJECT) objectid = serializer_writer.rint();
			else if (flags & SOUND_STATIC) {
				x_pos = serializer_writer.rfloat();
				y_pos = serializer_writer.rfloat();
			}
			int sound_handle = start_sound(new Sound_event(sample_name,flags,volume,x_pos,y_pos,objectid,this));
			delete [] sample_name;
			sound_handles_server->insert(std::pair<int,int>(server_handle,sound_handle));
		} else if (serializer_writer.next_int() == TT_SOUND_STOP) {
			serializer_writer.eat_int(TT_SOUND_STOP);
			int sound_handle = serializer_writer.rint();
			Sound_handles_server::iterator i = sound_handles_server->find(sound_handle);
			if (i != sound_handles_server->end())
				stop_sound(i->second);
		} else
			break;
	}

	update();
}

//-------------------------------------------------------------------
int Soundsystem::start_sound(Sound_event *sound_event)
{
	int sound_handle = next_sound_handle++;
	sound_events_handle->insert(std::pair<int, Sound_event*>(sound_handle,sound_event));
	return sound_handle;
}

//-------------------------------------------------------------------
int Soundsystem::start_sound(char *sample_name, int flags, float volume) // global
{
  logmsg ( lt_debug , "Playing sound %s" , sample_name );
	return start_sound(new Sound_event(sample_name,flags|SOUND_GLOBAL,volume,0,0,0,this));
}

int Soundsystem::start_sound(char *sample_name, int flags, float volume, float x_pos, float y_pos) // static
{
	return start_sound(new Sound_event(sample_name,flags|SOUND_STATIC,volume,x_pos,y_pos,0,this));
}

int Soundsystem::start_sound(char *sample_name, int flags, float volume, int objectid) // bound to object
{
	return start_sound(new Sound_event(sample_name,flags|SOUND_OBJECT,volume,0,0,objectid,this));
}

//-------------------------------------------------------------------
void Soundsystem::stop_sound(int sound_handle)
{
	Sound_events_handle::iterator i = sound_events_handle->find(sound_handle);
	if (i == sound_events_handle->end()) {
		logmsg(lt_debug,"Unable to stop non-existant sound with handle %d.",sound_handle);
		return;
	}
	i->second->bedone();
}
